/*******************************************************************************
 * Copyright 2015 - 2018 fmCreation
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald diese von der
 * Europischen Kommission genehmigt wurden - Folgeversionen der EUPL
 * ("Lizenz"); Sie drfen dieses Werk ausschlielich gem dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 *   https://progs.frankmuenster.de/eupl_v1-2_de
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften gefordert oder in
 * schriftlicher Form vereinbart, wird die unter der Lizenz verbreitete Software
 * "so wie sie ist", OHNE JEGLICHE GEWHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrcklich oder stillschweigend - verbreitet. Die sprachspezifischen
 * Genehmigungen und Beschrnkungen unter der Lizenz sind dem Lizenztext zu
 * entnehmen.
 ******************************************************************************/

package de.frankmuenster.mahoe.banks.santander.web;

import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.rmi.RemoteException;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Date;
import java.util.List;
import java.util.Locale;

import org.apache.commons.lang.StringUtils;

import com.gargoylesoftware.htmlunit.ElementNotFoundException;
import com.gargoylesoftware.htmlunit.FailingHttpStatusCodeException;
import com.gargoylesoftware.htmlunit.Page;
import com.gargoylesoftware.htmlunit.html.DomElement;
import com.gargoylesoftware.htmlunit.html.DomNode;
import com.gargoylesoftware.htmlunit.html.HtmlAnchor;
import com.gargoylesoftware.htmlunit.html.HtmlDivision;
import com.gargoylesoftware.htmlunit.html.HtmlElement;
import com.gargoylesoftware.htmlunit.html.HtmlPage;
import com.gargoylesoftware.htmlunit.html.HtmlTable;
import com.gargoylesoftware.htmlunit.html.HtmlTableBody;
import com.gargoylesoftware.htmlunit.html.HtmlTableRow;

import de.frankmuenster.mahoe.MahoeSynchronizeBackend;
import de.frankmuenster.mahoe.banks.santander.clearing.ClearingEntry;
import de.frankmuenster.mahoe.banks.santander.pdfextractor.SantanderPdfExtractor;
import de.frankmuenster.mahoe.pdfextractor.InvalidCardNumber;
import de.frankmuenster.mahoe.pdfextractor.PdfBuchung;
import de.frankmuenster.mahoe.utils.I18N;
import de.frankmuenster.mahoe.utils.TransactionSaver;
import de.frankmuenster.mahoe.utils.Utils;
import de.frankmuenster.mahoe.web.AbstractWebSite;
import de.frankmuenster.mahoe.web.WebSiteException;
import de.frankmuenster.mahoe.web.WebSiteExceptionTypes;
import de.willuhn.jameica.hbci.Settings;
import de.willuhn.jameica.hbci.rmi.HBCIDBService;
import de.willuhn.jameica.hbci.rmi.Konto;
import de.willuhn.jameica.hbci.rmi.KontoType;
import de.willuhn.jameica.hbci.rmi.Umsatz;
import de.willuhn.logging.Logger;
import de.willuhn.util.ApplicationException;

/**
 * Implementierung fr eine Kreditkarte.
 *
 * @author Frank Mnster
 */
@SuppressWarnings({ "PMD.ExcessiveImports", "PMD.GodClass" })
public class SantanderWebSiteKreditkarte
        extends AbstractSantanderWebSite<SantanderWebAccountInformation, SantanderBankMessages> {

    /**
     * Identifikationsstring f&uuml;r die Kreditakartentabelle
     */
    private static final String KREDITKARTEN = "Kreditkarten"; //$NON-NLS-1$

    /**
     * Text, der einen Systemfehler bei der Bank meldet.
     */
    private static final String SYS_ERROR_TEXT = "aufgrund eines Systemfehlers steht Ihnen zur Zeit das Internet Banking"; //$NON-NLS-1$

    /**
     * Die Ausgleichsbuchung
     */
    private ClearingEntry clearingEntry;

    /**
     * Marker fr den ersten Kontoauszug
     */
    private boolean firstPDF;

    /**
     * Indikator das letztes PDF erreicht ist.
     */
    private boolean lastPdfToScanHit;

    /**
     * Konstruktor fr einE Kreditkarte.
     *
     * @param konto
     * @param userName
     * @throws WebSiteException
     */
    protected SantanderWebSiteKreditkarte(final Konto konto, final String userName) throws WebSiteException {

        super(konto, userName);
    }

    /**
     * @see de.frankmuenster.mahoe.banks.santander.web.AbstractSantanderWebSite#getAccountInformation()
     */
    @Override
    public SantanderWebAccountInformation getAccountInformation() throws WebSiteException, ApplicationException {

        try {
            final String cardNumber = this.accountInfo.getKonto().getMeta(MahoeSynchronizeBackend.CARD_NUMBER,
                    StringUtils.EMPTY);
            // get anchor to details page
            final HtmlAnchor details = super.getAccountDetailsAnchor(cardNumber,
                    SantanderWebSiteKreditkarte.KREDITKARTEN, AbstractSantanderWebSite.ANCHOR_CELL_DETAILS);

            if (details == null) {
                throw new WebSiteException(I18N.tr("SantanderWebSiteKreditkarte.detailPageNotFound"), //$NON-NLS-1$
                        WebSiteExceptionTypes.PAGE_NOT_FOUND);
            } else {
                final HtmlPage detailsPage = details.click();
                Logger.debug(detailsPage.asXml());
                // Kartenart
                final HtmlElement kindOfCard = (HtmlElement) detailsPage
                        .getByXPath("/html/body/div/div[3]/div[1]/div/div[1]/div[1]/div[1]/span[2]").get(0); //$NON-NLS-1$
                this.accountInfo.setKindOfCard(kindOfCard.asText());
                // Limit
                final HtmlElement limit = (HtmlElement) detailsPage
                        .getByXPath("/html/body/div/div[3]/div[1]/div/div[1]/div[1]/div[5]/span[2]").get(0); //$NON-NLS-1$
                final StringBuilder comment = new StringBuilder(I18N.tr("SantanderWebSiteKreditkarte.limitText")); //$NON-NLS-1$
                comment.append(limit.asText());
                this.accountInfo.setComment(comment.toString());
                // Saldo
                final HtmlElement saldo = (HtmlElement) detailsPage
                        .getByXPath("/html/body/div/div[3]/div[1]/div/div[1]/div[1]/div[6]/span[2]").get(0); //$NON-NLS-1$
                this.accountInfo.setSaldo(Utils.convertStringToBigDecimal(saldo.asText()));
            }
        } catch (FailingHttpStatusCodeException | IOException e) {
            Logger.error(I18N.tr("SantanderWebSiteKreditkarte.readAccountOverviewFailedLog", //$NON-NLS-1$
                    e.getMessage()), e);
            throw new WebSiteException(I18N.tr("SantanderWebSiteKreditkarte.readAccountOverviewFailed"), // NOPMD //$NON-NLS-1$
                    WebSiteExceptionTypes.READ_ACCOUNT_DETAILS_FAILED, e);
        }
        return this.accountInfo;
    }

    /**
     * @see de.frankmuenster.mahoe.banks.santander.web.AbstractSantanderWebSite#getBookedTransactions()
     */
    @Override
    public List<Umsatz> getBookedTransactions() throws WebSiteException, ApplicationException {

        List<Umsatz> transactions = new ArrayList<>();
        try {
            final HtmlPage filteredPage = SantanderWebSiteUtils.getDocumentFilteredPage(this.webClient,
                    DocumentType.KARTENABRECHNUNG);
            if (filteredPage.asText().contains("Es liegen zur Zeit keine Nachrichten fr Sie vor")) { //$NON-NLS-1$
                Logger.info(I18N.tr("SantanderWebSiteKreditkarte.noCreditCardInvoice")); //$NON-NLS-1$
            } else {
                transactions = this.getTransactionsFromPdfs(filteredPage);
            }
        } catch (FailingHttpStatusCodeException | IOException e) {
            Logger.error(I18N.tr("openPdfOviewFailedLog", e.getMessage()), e); //$NON-NLS-1$
            throw new WebSiteException(I18N.tr("openPdfOviewFailed"), //$NON-NLS-1$
                    WebSiteExceptionTypes.READ_ACCOUNT_DETAILS_FAILED, e);
        }
        return transactions;
    }

    /**
     * Liefert die Ausgleichsbuchung
     *
     * @return die Ausgleichsbuchung
     */
    public ClearingEntry getClearingEntry() {

        return this.clearingEntry;
    }

    /**
     * @see de.frankmuenster.mahoe.banks.santander.web.AbstractSantanderWebSite#getPreBookedTransactions()
     */
    @Override
    public List<Umsatz> getPreBookedTransactions() throws WebSiteException, ApplicationException {

        List<Umsatz> transactions = new ArrayList<>();
        try {

            final String cardNumber = this.accountInfo.getKonto().getMeta(MahoeSynchronizeBackend.CARD_NUMBER,
                    StringUtils.EMPTY);

            // get transactions anchor
            final HtmlAnchor anchor = super.getAccountDetailsAnchor(cardNumber,
                    SantanderWebSiteKreditkarte.KREDITKARTEN, AbstractSantanderWebSite.ANCHOR_CELL_TRANSACTIONS);

            if (anchor == null) {
                throw new WebSiteException(I18N.tr("SantanderWebSiteKreditkarte.preBookedPageNotFound"), //$NON-NLS-1$
                        WebSiteExceptionTypes.PAGE_NOT_FOUND);
            } else {
                transactions = this.getPreBookedTransactions(anchor);
            }
        } catch (FailingHttpStatusCodeException | IOException e) {
            Logger.error(I18N.tr("SantanderWebSiteKreditkarte.getPreBookedOverviewFailedLog", //$NON-NLS-1$
                    e.getMessage()), e);
            throw new WebSiteException(I18N.tr("SantanderWebSiteKreditkarte.getPreBookedOverviewFailed"), // NOPMD //$NON-NLS-1$
                    WebSiteExceptionTypes.READ_ACCOUNT_DETAILS_FAILED, e);
        }
        return transactions;
    }

    /**
     * @see de.frankmuenster.mahoe.banks.santander.web.AbstractSantanderWebSite#getStartdate()
     */
    @Override
    protected Calendar getStartdate() {

        // Setzte das startDatum auf jetzt minus 360 Tage
        final Calendar startCalendar = Calendar.getInstance();
        startCalendar.add(Calendar.DAY_OF_YEAR, -720);
        return startCalendar;
    }

    /**
     * @see de.frankmuenster.mahoe.banks.santander.web.AbstractSantanderWebSite#isSupporting(de.willuhn.jameica.hbci.rmi.KontoType)
     */
    @Override
    protected boolean isSupporting(final KontoType type) {

        return KontoType.KREDITKARTE.equals(type);
    }

    /**
     * Liefert eine Ausgleichbuchung aus den Daten des Auszuges
     *
     * @param extractor
     *            der PDFExtraktor
     */
    private void createClearingEntry(final SantanderPdfExtractor extractor) {

        // Ausgleichsbuchung anlegen, wenn erster Kontoauszug
        if (this.firstPDF) {
            this.clearingEntry = extractor.getAusgleichsBuchung();
            this.firstPDF = false;
        }
    }

    private HtmlPage getNextPage(final HtmlPage detailsPage) {

        HtmlPage nextPage = null;

        // Suche nach weiteren Seiten
        if (!this.lastPdfToScanHit) {
            try {
                final HtmlAnchor nextLink = detailsPage.getAnchorByText(AbstractSantanderWebSite.NEXT_LINK_TEXT);
                nextPage = nextLink.click();
            } catch (final ElementNotFoundException e) {
                Logger.debug(I18N.tr(AbstractSantanderWebSite.I18N_NEXT_LINK_NOT_FOUND));
                this.lastPdfToScanHit = true;
            } catch (final IOException e) {
                Logger.error(I18N.tr("SantanderWebSiteKreditkarte.loadNextPageFailed"), e); //$NON-NLS-1$
                this.lastPdfToScanHit = true;
            }
        }

        return nextPage;
    }

    /**
     * @param anchor
     * @return eine Liste mit den gefundenen Ums&auml;tzen.
     * @throws IOException
     */
    private List<Umsatz> getPreBookedTransactions(final HtmlAnchor anchor) throws IOException {

        final List<Umsatz> transactions = new ArrayList<>();
        HtmlAnchor nextLink = anchor;
        do {
            final HtmlPage detailsPage = nextLink.click();
            Logger.debug(detailsPage.asXml());
            // Testen, ob Systemfehler angezeigt wird
            if (detailsPage.asText().contains(SantanderWebSiteKreditkarte.SYS_ERROR_TEXT)) {
                Logger.info(I18N.tr("SantanderWebSiteKreditkarte.bankSystemError")); //$NON-NLS-1$
                throw new IOException(I18N.tr("SantanderWebSiteKreditkarte.bankSystemError")); //$NON-NLS-1$
            }
            final List<DomElement> tables = detailsPage.getElementsByTagName(AbstractWebSite.TABLE_TAG);
            HtmlTable detailsTable = null;
            for (final DomElement domElement : tables) {
                detailsTable = (HtmlTable) domElement;
                Logger.debug(detailsTable.getFirstChild().getNextSibling().getTextContent());
                if ("Transactions".equals(detailsTable.getFirstChild().getNextSibling().getTextContent())) { //$NON-NLS-1$
                    transactions.addAll(this.readTransactionfromTable(detailsTable));
                }
            }
            // Suche nach weiteren Seiten
            try {
                nextLink = detailsPage.getAnchorByText(AbstractSantanderWebSite.NEXT_LINK_TEXT);
            } catch (final ElementNotFoundException e) {
                Logger.debug(I18N.tr(AbstractSantanderWebSite.I18N_NEXT_LINK_NOT_FOUND));
                break;
            }
        } while (true);
        return transactions;
    }

    /**
     * @param row
     * @param cardNumber
     * @return the List<Umsatz> with found transactions in the table row
     * @throws RemoteException
     */
    private List<Umsatz> getTransactionsFromPdfInRow(final HtmlTableRow row, final String cardNumber)
            throws RemoteException {

        final SimpleDateFormat sdf = new SimpleDateFormat("dd.MM.yy", Locale.GERMAN); //$NON-NLS-1$
        final List<Umsatz> transactions = new ArrayList<>();
        final Date startDate = this.accountInfo.getLastTransactionDate();
        final String pdfDateString = row.getCell(3).getTextContent().trim().replaceAll("\u00a0", //$NON-NLS-1$
                StringUtils.EMPTY);
        Logger.debug(I18N.tr("SantanderWebSiteKreditkarte.billingDate", pdfDateString)); //$NON-NLS-1$
        try {
            final Date pdfDate = sdf.parse(pdfDateString);
            if (pdfDate.before(startDate)) {
                Logger.debug(I18N.tr("TargoBankWebSiteTransactionsImpl.reachedLastPdf")); //$NON-NLS-1$
                this.lastPdfToScanHit = true;
            }
        } catch (final ParseException e1) {
            Logger.error(I18N.tr("TargoBankWebSiteTransactionsImpl.readPdfDateFailed", e1.getMessage()), //$NON-NLS-1$
                    e1);
        }
        final DomNode urlText = row.getCell(1).getFirstChild().getNextSibling();
        Logger.debug("URLtext: " + urlText); //$NON-NLS-1$
        HtmlAnchor anchor = null;
        if (urlText instanceof HtmlAnchor) {
            anchor = (HtmlAnchor) urlText;
            Logger.debug("Abrechnung gefunden."); //$NON-NLS-1$
            try {
                final URL fullUrl = row.getHtmlPageOrNull().getFullyQualifiedUrl(anchor.getHrefAttribute());
                anchor.setAttribute("href", fullUrl.toString()); //$NON-NLS-1$
                Logger.debug("URL: " + anchor.getHrefAttribute()); //$NON-NLS-1$
                // final List<Umsatz> transactions =
                transactions.addAll(this.loadTransactionsFromPDF(anchor, cardNumber));
            } catch (final MalformedURLException e) {
                Logger.error(I18N.tr("TargoBankWebSiteTransactionsImpl.pdfUrlMalformend", //$NON-NLS-1$
                        e.getMessage()), e);
            }
        } else {
            Logger.error(I18N.tr("SantanderWebSiteKreditkarte.pdfUrlNotFound")); //$NON-NLS-1$
        }
        return transactions;
    }

    /**
     * Liefert eine Liste mit den Ums&auml;tzen aus den PDF der Seite
     *
     * @param page
     *            to inspect
     * @return Liste mit den gefundenen Ums&auml;tzen
     * @throws RemoteException
     */
    private List<Umsatz> getTransactionsFromPdfs(final HtmlPage page) throws RemoteException {

        final List<Umsatz> transactions = new ArrayList<>();
        final String cardNumber = this.accountInfo.getKonto().getMeta(MahoeSynchronizeBackend.CARD_NUMBER,
                StringUtils.EMPTY);
        HtmlPage tablePage = page;
        this.lastPdfToScanHit = false;
        this.firstPDF = true;
        do {
            final HtmlDivision content = (HtmlDivision) tablePage.getElementById("content"); //$NON-NLS-1$
            final List<HtmlTable> pdfTables = content.getElementsByAttribute("table", "summary", "List of payments"); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
            if (pdfTables.size() == AbstractSantanderWebSite.MAX_TABLE_SIZE) {
                final HtmlTableBody body = pdfTables.get(0).getBodies().get(0);
                for (final HtmlTableRow row : body.getRows()) {
                    if (row.getCell(2).getTextContent().contains(cardNumber)) {
                        transactions.addAll(this.getTransactionsFromPdfInRow(row, cardNumber));
                        if (this.lastPdfToScanHit) {
                            break;
                        }
                    }
                }
            }
            tablePage = this.getNextPage(tablePage);
        } while (tablePage != null);

        return transactions;
    }

    private Umsatz loadTransaction(final HtmlPage transactionPage) throws RemoteException {

        final HBCIDBService dbService = Settings.getDBService();
        final SimpleDateFormat sdf = new SimpleDateFormat("dd.MM.yyyy", //$NON-NLS-1$
                Locale.getDefault());
        Logger.debug(transactionPage.asText());
        Umsatz umsatz = null;
        // Datum
        final HtmlElement dateField = (HtmlElement) transactionPage
                .getByXPath("//div[@class='contents']/div/div[1]/span[2]").get(0); //$NON-NLS-1$
        // Ist Datum ein leerer String, dann abbrechen
        if (StringUtils.isNotEmpty(dateField.asText().trim())) {
            umsatz = dbService.createObject(Umsatz.class, null);
            try {
                final Date date = sdf.parse(dateField.asText());
                umsatz.setDatum(date);
                umsatz.setValuta(date);
            } catch (final ParseException e) {
                Logger.error(I18N.tr("SantanderWebSiteKreditkarte.parseBookingDateFailed", e.getMessage()), //$NON-NLS-1$
                        e);
            }
            // Art der Transaktion
            final HtmlElement typField = (HtmlElement) transactionPage
                    .getByXPath("//div[@class='contents']/div/div[3]/span[2]").get(0); //$NON-NLS-1$
            umsatz.setArt(typField.asText());
            // Einreicher --> Zweck
            final HtmlElement requestorField = (HtmlElement) transactionPage
                    .getByXPath("//div[@class='contents']/div/div[4]/span[2]").get(0); //$NON-NLS-1$
            final HtmlElement placeField = (HtmlElement) transactionPage
                    .getByXPath("//div[@class='contents']/div/div[5]/span[2]").get(0); //$NON-NLS-1$
            final StringBuilder purpose = new StringBuilder(requestorField.asText()).append('\n');
            purpose.append(placeField.asText());
            if (purpose.length() == 1) { // NOPMD
                // Keine Zweck gefunden --> Art speichern, falls vorhanden
                if (StringUtils.isNotEmpty(umsatz.getArt())) {
                    Utils.loadUmsatz(umsatz, umsatz.getArt());
                }
            } else {
                Utils.loadUmsatz(umsatz, purpose.toString());
            }
            // Betrag
            final HtmlElement amountField = (HtmlElement) transactionPage
                    .getByXPath("//div[@class='contents']/div/div[6]/span[2]").get(0); //$NON-NLS-1$
            umsatz.setBetrag(Utils.convertStringToDouble(amountField.asText()));
            // Fix Tankrabatt
            if ("Ihr Tankrabatt".equals(umsatz.getArt())) { //$NON-NLS-1$
                umsatz.setBetrag(Math.abs(umsatz.getBetrag()));
            }
            // Auslndische Whrung --> Kommentar
            final HtmlElement foreignAmmontFld = (HtmlElement) transactionPage
                    .getByXPath("//div[@class='contents']/div/div[8]/span[2]").get(0); //$NON-NLS-1$
            if (StringUtils.isNotEmpty(foreignAmmontFld.asText())) {
                final HtmlElement currencyFld = (HtmlElement) transactionPage
                        .getByXPath("//div[@class='contents']/div/div[7]/span[2]").get(0); //$NON-NLS-1$
                final HtmlElement exchRateFld = (HtmlElement) transactionPage
                        .getByXPath("//div[@class='contents']/div/div[9]/span[2]").get(0); //$NON-NLS-1$
                final StringBuilder comment = new StringBuilder(I18N.tr("SantanderWebSiteKreditkarte.foreignAmount")) //$NON-NLS-1$
                        .append(foreignAmmontFld.asText()).append(' ').append(currencyFld.asText())
                        .append(I18N.tr("SantanderWebSiteKreditkarte.exchangeRate")).append(exchRateFld.asText()); //$NON-NLS-1$
                umsatz.setKommentar(comment.toString());
            }
            umsatz.setFlags(Umsatz.FLAG_NOTBOOKED);
        }
        return umsatz;
    }

    /**
     * Lde alle Umstze aus dem im Anchor angegebenen PDF
     *
     * @param anchor
     *            zu dem PDF.
     * @param cardNumber
     * @return Liste mit allen gefundenen Umstzen. Kann leer sein.
     */
    private List<Umsatz> loadTransactionsFromPDF(final HtmlAnchor anchor, final String cardNumber) {

        List<Umsatz> transactions = new ArrayList<>();
        InputStream stream = null;
        try {
            final Page pdf = this.webClient.getPage(anchor.getHrefAttribute());
            if (pdf.isHtmlPage()) {
                Logger.debug("PDF: " + ((HtmlPage) pdf).asText()); //$NON-NLS-1$
                Logger.error(I18N.tr("SantanderWebSiteKreditkarte.loadPdfFailed", anchor.getHrefAttribute())); //$NON-NLS-1$
            } else {
                Logger.debug("Lade PDF"); //$NON-NLS-1$
                stream = pdf.getWebResponse().getContentAsStream();
                final SantanderPdfExtractor extractor = new SantanderPdfExtractor(stream, cardNumber);
                Logger.debug(I18N.tr("SantanderWebSiteKreditkarte.extractTransactions", cardNumber)); //$NON-NLS-1$
                final List<PdfBuchung> buchungenPDF = extractor.getExtract();
                // Ausgleichsbuchung
                this.createClearingEntry(extractor);
                // Ressourcen schlieen
                extractor.close();
                stream.close();
                transactions = TransactionSaver.getTransactions(buchungenPDF);
            }
        } catch (IOException | InvalidCardNumber e) {
            Logger.error(I18N.tr("TargoBankWebSiteTransactionsImpl.readPdfContentFailed", e.getMessage()), e); //$NON-NLS-1$
            if (stream != null) {
                try {
                    stream.close();
                } catch (final IOException e1) { // NOPMD
                    // Nothing to do here
                }
            }
        }
        return transactions;
    }

    /**
     * Loop through the HTML table with pre-booked transactions and read transactions from each row.
     *
     * @param detailsTable
     *            with pre-booked transactions
     * @return a list of found transactions
     * @throws IOException
     */
    private List<Umsatz> readTransactionfromTable(final HtmlTable detailsTable) throws IOException {

        final List<Umsatz> transactions = new ArrayList<>();

        final List<HtmlTableRow> rows = detailsTable.getBodies().get(0).getRows();
        for (final HtmlTableRow row : rows) {
            final HtmlAnchor transAnchor = (HtmlAnchor) row.getCell(0).getFirstElementChild();
            final HtmlPage transactionPage = transAnchor.click();
            final Umsatz transaction = this.loadTransaction(transactionPage);
            if (transaction != null) {
                transactions.add(transaction);
            }
        }

        return transactions;

    }

}
