/*******************************************************************************
 * Copyright 2015 - 2018 fmCreation
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald diese von der
 * Europischen Kommission genehmigt wurden - Folgeversionen der EUPL
 * ("Lizenz"); Sie drfen dieses Werk ausschlielich gem dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 *   https://progs.frankmuenster.de/eupl_v1-2_de
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften gefordert oder in
 * schriftlicher Form vereinbart, wird die unter der Lizenz verbreitete Software
 * "so wie sie ist", OHNE JEGLICHE GEWHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrcklich oder stillschweigend - verbreitet. Die sprachspezifischen
 * Genehmigungen und Beschrnkungen unter der Lizenz sind dem Lizenztext zu
 * entnehmen.
 ******************************************************************************/

package de.frankmuenster.mahoe.banks.targobank.web;

import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.rmi.RemoteException;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.List;
import java.util.Locale;

import org.apache.commons.lang.StringUtils;

import com.gargoylesoftware.htmlunit.FailingHttpStatusCodeException;
import com.gargoylesoftware.htmlunit.WebClient;
import com.gargoylesoftware.htmlunit.html.DomNode;
import com.gargoylesoftware.htmlunit.html.DomNodeList;
import com.gargoylesoftware.htmlunit.html.HtmlAnchor;
import com.gargoylesoftware.htmlunit.html.HtmlElement;
import com.gargoylesoftware.htmlunit.html.HtmlEmphasis;
import com.gargoylesoftware.htmlunit.html.HtmlInput;
import com.gargoylesoftware.htmlunit.html.HtmlOption;
import com.gargoylesoftware.htmlunit.html.HtmlPage;
import com.gargoylesoftware.htmlunit.html.HtmlSelect;
import com.gargoylesoftware.htmlunit.html.HtmlTable;
import com.gargoylesoftware.htmlunit.html.HtmlTableBody;
import com.gargoylesoftware.htmlunit.html.HtmlTableRow;

import de.frankmuenster.mahoe.banks.targobank.TargoBank;
import de.frankmuenster.mahoe.statements.AbstractAccountStatementFinder;
import de.frankmuenster.mahoe.statements.MahoeAccountStatement;
import de.frankmuenster.mahoe.utils.I18N;
import de.frankmuenster.mahoe.utils.WebUtils;
import de.frankmuenster.mahoe.web.AbstractWebSite;
import de.frankmuenster.mahoe.web.WebSiteException;
import de.frankmuenster.mahoe.web.WebSiteExceptionTypes;
import de.willuhn.logging.Logger;

/**
 * Class to get account statements from the banks web page
 *
 * @author Frank Mnster
 */
@SuppressWarnings({ "PMD.ExcessiveImports" })
public class TargoBankAccountStatementFinder extends AbstractAccountStatementFinder<TargoBankWebAccountInformation> {

    /**
     * Minimum cell size of a table row with account sheet PDFs
     */
    private static final int MIN_ROW_CELL_SIZE = 2;

    /**
     * Creates a new TargoBankAccountStatementFinder object with a webClient and
     * account information.
     *
     * @param webClient
     * @param accountInfo
     * @throws WebSiteException
     */
    public TargoBankAccountStatementFinder(final WebClient webClient, final TargoBankWebAccountInformation accountInfo)
            throws WebSiteException {

        super(webClient, accountInfo);

    }

    /**
     * @throws WebSiteException
     * @see AbstractAccountStatementFinder#startSearch()
     */
    @Override
    protected void startSearch() throws WebSiteException {

        HtmlPage pageAbr;
        try {
            pageAbr = this.webClient.getPage(TargoBank.URL_ABBRECHNUNGEN);
            if (pageAbr.asText().contains("Online-Postfach")) { //$NON-NLS-1$

                // Choose correct account via credit card number
                final String kartenId = "TARGO   0000" + this.accountInfo.getCardAccount(); //$NON-NLS-1$
                final HtmlSelect selIndex1 = pageAbr.getElementByName("SelIndex1"); //$NON-NLS-1$
                final HtmlInput submit = (HtmlInput) pageAbr.getElementById("btnRechercher"); //$NON-NLS-1$
                final HtmlOption option = selIndex1.getOptionByValue(kartenId);
                selIndex1.setSelectedAttribute(option, true);
                // OPen card details
                HtmlPage pageNew = submit.click();
                // indicator for NEXT page link
                boolean hasNext;

                do {
                    this.processStatementPage(pageNew);

                    hasNext = false;
                    final HtmlTable table = (HtmlTable) pageNew.getElementById("Table1"); //$NON-NLS-1$
                    final DomNodeList<HtmlElement> inputButtons = table.getElementsByTagName(HtmlInput.TAG_NAME);
                    for (final HtmlElement htmlElement : inputButtons) {
                        if (htmlElement.getId().equals("NEXT")) { //$NON-NLS-1$
                            hasNext = true;
                            pageNew = htmlElement.click();
                            break;
                        }
                    }
                } while (hasNext);
            } else {
                Logger.warn(TargoBankWebSite.PDFSITE_NOT_FOUND);
                throw new WebSiteException(I18N.tr("AccountStatementFinder.downloadFailed"), //$NON-NLS-1$
                        WebSiteExceptionTypes.PAGE_NOT_FOUND);
            }
        } catch (FailingHttpStatusCodeException | IOException e) {
            Logger.error(I18N.tr("TargoBankWebSite.loadPdfFailedLog", e.getMessage()), e); //$NON-NLS-1$
            throw new WebSiteException(I18N.tr("AccountStatementFinder.downloadFailed"), //$NON-NLS-1$
                    WebSiteExceptionTypes.PAGE_NOT_FOUND, e);
        }

    }

    /**
     * Process the statements on the actual page and checks for additional pages
     * as well.
     *
     * @param page
     *            the first page to start processing with
     */
    private void processStatementPage(final HtmlPage page) {

        final SimpleDateFormat sdf = new SimpleDateFormat("dd.MM.yyyy", Locale.GERMAN); //$NON-NLS-1$
        final HtmlTable table = (HtmlTable) page.getElementById("Table1"); //$NON-NLS-1$
        final DomNodeList<HtmlElement> tables = table.getElementsByTagName(HtmlTable.TAG_NAME);
        final HtmlTableBody tbBody = ((HtmlTable) tables.item(1)).getBodies().get(0);
        final List<HtmlTableRow> tbRows = tbBody.getRows();
        for (final HtmlTableRow tbRow : tbRows) {
            if (tbRow.getCells().size() < TargoBankAccountStatementFinder.MIN_ROW_CELL_SIZE) {
                continue;
            }

            // Table row seems to contain a PDF sheet --> test and save
            this.saveSinglePdfSheet(tbRow, sdf);

        } // end for tbRows

    }

    /**
     * Checks the given table row if a PDF sheet anchor exists and downloads it.
     *
     * @param tbRow
     *            the table row
     * @param sdf
     *            {@link SimpleDateFormat} of date in table row
     */
    private void saveSinglePdfSheet(final HtmlTableRow tbRow, final SimpleDateFormat sdf) {

        final String pdfDateString = tbRow.getCell(1).getTextContent().trim()
                .replaceAll(AbstractWebSite.NON_BREAK_SPACE, StringUtils.EMPTY);
        Logger.debug("Abrechnungsdatum: " + pdfDateString); //$NON-NLS-1$
        try {
            final Date pdfDate = sdf.parse(pdfDateString);
            final DomNode urlText = tbRow.getCell(2).getFirstChild();
            Logger.debug("URLtext: " + urlText); //$NON-NLS-1$
            HtmlAnchor anchor = null;
            if (urlText instanceof HtmlAnchor) {
                anchor = (HtmlAnchor) urlText;
                Logger.debug("Alte Abrechnung gefunden."); //$NON-NLS-1$
            } else if (urlText instanceof HtmlEmphasis) {
                Logger.debug("Neue Abrechnung gefunden."); //$NON-NLS-1$
                anchor = (HtmlAnchor) urlText.getFirstChild();
            } else {
                Logger.error(I18N.tr("TargoBankWebSite.getPdfUrlFailed")); //$NON-NLS-1$
            }
            if (anchor != null) {
                final URL fullUrl = tbRow.getHtmlPageOrNull().getFullyQualifiedUrl(anchor.getHrefAttribute());
                anchor.setAttribute("href", fullUrl.toString()); //$NON-NLS-1$
                Logger.debug("URL: " + anchor.getHrefAttribute()); //$NON-NLS-1$

                // Download PDF as byte array
                final byte[] data = WebUtils.getPdfAsByteArray(anchor);
                // create an MahoeAccountStatement
                final MahoeAccountStatement statementInfo = new MahoeAccountStatement(pdfDate, TargoBank.NAME);
                // statement to internal list
                this.addStatement(statementInfo, data);

            }
        } catch (final MalformedURLException e) {
            Logger.error(I18N.tr("TargoBankWebSite.pdfUrlMalforment", e.getMessage()), e); //$NON-NLS-1$
        } catch (final ParseException e) {
            Logger.error(I18N.tr("TargoBankWebSite.parsePdfDateFailed", e.getMessage()), e); //$NON-NLS-1$
        } catch (final RemoteException e) {
            Logger.error("PDF konnte nicht gespeichert werden. Fehler: {0}.", e); //$NON-NLS-1$
        }

    }

}
