/*******************************************************************************
 * Copyright 2015 - 2018 fmCreation
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald diese von der
 * Europischen Kommission genehmigt wurden - Folgeversionen der EUPL
 * ("Lizenz"); Sie drfen dieses Werk ausschlielich gem dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 *   https://progs.frankmuenster.de/eupl_v1-2_de
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften gefordert oder in
 * schriftlicher Form vereinbart, wird die unter der Lizenz verbreitete Software
 * "so wie sie ist", OHNE JEGLICHE GEWHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrcklich oder stillschweigend - verbreitet. Die sprachspezifischen
 * Genehmigungen und Beschrnkungen unter der Lizenz sind dem Lizenztext zu
 * entnehmen.
 ******************************************************************************/

package de.frankmuenster.mahoe.jobs;

import java.rmi.RemoteException;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;

import de.frankmuenster.mahoe.utils.I18N;
import de.frankmuenster.mahoe.web.MahoeKontoauszugPDF;
import de.frankmuenster.mahoe.web.WebAccountInformation;
import de.frankmuenster.mahoe.web.WebSite;
import de.frankmuenster.mahoe.web.WebSiteException;
import de.willuhn.datasource.GenericIterator;
import de.willuhn.jameica.hbci.HBCI;
import de.willuhn.jameica.hbci.MetaKey;
import de.willuhn.jameica.hbci.messaging.ImportMessage;
import de.willuhn.jameica.hbci.rmi.Konto;
import de.willuhn.jameica.hbci.rmi.Kontoauszug;
import de.willuhn.jameica.hbci.rmi.Protokoll;
import de.willuhn.jameica.hbci.server.KontoauszugPdfUtil;
import de.willuhn.jameica.hbci.synchronize.jobs.SynchronizeJob;
import de.willuhn.jameica.hbci.synchronize.jobs.SynchronizeJobKontoauszugPdf;
import de.willuhn.jameica.system.Application;
import de.willuhn.logging.Logger;
import de.willuhn.util.ApplicationException;
import de.willuhn.util.ProgressMonitor;

/**
 * Abstrakte Klasse fr Kontoauszugsabfrage PDF fr Mahoe Konten. Hier verarbeiten wir hauptschlich die
 * Fehlerbehandlung.
 *
 * @author Frank Mnster
 * @param <I>
 *            die bankspezifischen WebAccountInformation
 */
public abstract class AbstractMahoeSynchronizeJobKontoauszugPdf<I extends WebAccountInformation>
        extends SynchronizeJobKontoauszugPdf implements InterfaceMahoeSynchronizeJob {

    /**
     * Das Konto
     */
    protected transient Konto konto;

    /**
     * Die Kundennummer. Extra damit man nicht jedes mal die RemoteException abfangen muss.
     */
    protected transient String kundennummer;

    /**
     * Das Webseiten Objekt
     */
    protected transient WebSite<I> webSite;

    /**
     * Kontoinformationen
     */
    protected transient I accountInfo;

    /**
     * Das Progessmonitor Objekt
     */
    protected transient ProgressMonitor monitor;

    /**
     * Checks if the given account statement already exists in the account.
     *
     * @param newStatement
     *            to be checked
     * @return <i>true</i> if the statement exists already
     */
    private static boolean isDoubleEntry(final Kontoauszug newStatement) {

        boolean isDouble = false;
        try {

            // get account from statement
            final Konto account = newStatement.getKonto();

            // get statement list from account
            final GenericIterator<Kontoauszug> existingStatemets = KontoauszugPdfUtil.getList(account,
                    newStatement.getErstellungsdatum(), newStatement.getErstellungsdatum(), false);

            // Loop over all existing statements until one equal has been found
            // or the end is reached
            while (existingStatemets.hasNext() && !isDouble) {
                final Kontoauszug storedStatemant = existingStatemets.next();
                isDouble = AbstractMahoeSynchronizeJobKontoauszugPdf.statementsEqual(newStatement, storedStatemant);
            }
        } catch (final RemoteException e) {
            Logger.warn(I18N.tr("TransactionSaver.isDoubleCheckFailed") + e.getMessage()); //$NON-NLS-1$
        }
        return isDouble;
    }

    /**
     * Check if two Objects are equal with null checking.
     *
     * @param obj1
     *            first object
     * @param obj2
     *            second object
     * @return <i>true</i> if this object is the same as the obj argument; <i>false</i> otherwise.
     */
    private static boolean objectEqual(final Object obj1, final Object obj2) {

        if (obj1 == null && obj2 == null) {
            return true; // NOPMD
        }
        return obj1 != null && obj1.equals(obj2);
    }

    /**
     * Compares to statements and returns <i>true</i> in case most relevant fields are equal, <i>false</i> else.
     *
     * @param statement1
     *            first statement
     * @param statemant2
     *            second statement
     * @return <i>true</i> if statements are considered to be equal, <i>false</i> else.
     * @throws RemoteException
     *             in case an attributes of an statement cannot be read from the database object.
     */
    private static boolean statementsEqual(final Kontoauszug statement1, final Kontoauszug statemant2)
            throws RemoteException {

        if (!AbstractMahoeSynchronizeJobKontoauszugPdf.objectEqual(statemant2.getKommentar(),
                statement1.getKommentar())) {
            return false; // NOPMD
        }

        if (!AbstractMahoeSynchronizeJobKontoauszugPdf.objectEqual(statemant2.getName1(), statement1.getName1())) {
            return false; // NOPMD
        }

        if (!AbstractMahoeSynchronizeJobKontoauszugPdf.objectEqual(statemant2.getName2(), statement1.getName2())) {
            return false; // NOPMD
        }

        if (!AbstractMahoeSynchronizeJobKontoauszugPdf.objectEqual(statemant2.getName3(), statement1.getName3())) {
            return false; // NOPMD
        }

        if (!AbstractMahoeSynchronizeJobKontoauszugPdf.objectEqual(statemant2.getErstellungsdatum(),
                statement1.getErstellungsdatum())) {
            return false; // NOPMD
        }

        if (!AbstractMahoeSynchronizeJobKontoauszugPdf.objectEqual(statemant2.getVon(), statement1.getVon())) {
            return false; // NOPMD
        }

        if (!AbstractMahoeSynchronizeJobKontoauszugPdf.objectEqual(statemant2.getBis(), statement1.getBis())) {
            return false; // NOPMD
        }

        if (!AbstractMahoeSynchronizeJobKontoauszugPdf.objectEqual(statemant2.getJahr(), statement1.getJahr())) {
            return false; // NOPMD
        }

        // Until now all equal. Simply return result from last check...
        return AbstractMahoeSynchronizeJobKontoauszugPdf.objectEqual(statemant2.getNummer(), statement1.getNummer());
    }

    /**
     * Standardkonstruktor
     */
    public AbstractMahoeSynchronizeJobKontoauszugPdf() {

        super();
    }

    /**
     * @see InterfaceMahoeSynchronizeJob#execute(ProgressMonitor, boolean)
     */
    @Override
    public void execute(final ProgressMonitor monitor, final boolean lastJob) throws Exception { // NOPMD

        this.konto = (Konto) this.getContext(SynchronizeJob.CTX_ENTITY);
        this.monitor = monitor;

        String accountname;
        try {
            this.kundennummer = this.konto.getKundennummer();
            accountname = this.konto.getBezeichnung();
        } catch (final RemoteException e) {
            throw new ApplicationException(I18N.tr("AbstractMahoeSynchronizeJobKontoauszug.readAccountInfoFailed"), e); //$NON-NLS-1$
        }
        Logger.info(I18N.tr("AbstractMahoeSynchronizeJobKontoauszug.readTransactionLoginfo", accountname, //$NON-NLS-1$
                this.kundennummer));

        // Wir laden das BankSeiten-Objekt
        this.getWebSite();
        monitor.setPercentComplete(10);

        // Nur Anmelden, wenn die Verbindung nicht schon besteht
        if (this.webSite.isStillLoggedIn()) {
            monitor.setStatusText(I18N.tr("AbstractMahoeSynchronizeJobKontoauszug.connectionStillValid")); //$NON-NLS-1$
        } else {
            monitor.setStatusText(I18N.tr("AbstractMahoeSynchronizeJobKontoauszug.startLogin")); //$NON-NLS-1$
            this.login();
            Logger.info(I18N.tr("AbstractMahoeSynchronizeJobKontoauszug.loginSuccessful")); //$NON-NLS-1$
            monitor.setStatusText(I18N.tr("AbstractMahoeSynchronizeJobKontoauszug.loginSuccessful")); //$NON-NLS-1$
        }
        monitor.setPercentComplete(20);

        // Kontoinformationen
        monitor.setStatusText(I18N.tr("AbstractMahoeSynchronizeJobKontoauszug.updateAccountInfo")); //$NON-NLS-1$
        this.accountInfo = this.webSite.getAccountInformation();
        monitor.setPercentComplete(30);

        // Kontoauszge
        monitor.setStatusText(I18N.tr("AbstractMahoeSynchronizeJobKontoauszug.readAccountSheets")); //$NON-NLS-1$
        this.handleAccountSheets();
        monitor.setPercentComplete(80);

        // Falls letzter Job fr die Bank, abmelden
        if (lastJob) {
            monitor.setStatusText(I18N.tr("AbstractMahoeSynchronizeJobKontoauszug.reachedLastJob")); //$NON-NLS-1$
            this.webSite.logoff();
        }
        monitor.setPercentComplete(100);

    }

    /**
     * Liefert das Webseiten-Objekt fr die Bank.
     *
     * @throws WebSiteException
     * @throws ApplicationException
     */
    protected abstract void getWebSite() throws WebSiteException, ApplicationException;

    /**
     * Speichert die Kontoauszge als PDF.
     *
     * @throws WebSiteException
     * @throws ApplicationException
     */
    protected void handleAccountSheets() throws WebSiteException, ApplicationException {

        try {

            final List<MahoeKontoauszugPDF> accSheets = this.getAccoutStatements();
            final Date now = new Date();

            for (final MahoeKontoauszugPDF kontoauszugPDF : accSheets) {

                final Kontoauszug ktoauszug = kontoauszugPDF.getAccountStatement();
                Logger.info(I18N.tr("AbstractMahoeSynchronizeJobKontoauszugPdf.saveAccountStatementInfo", //$NON-NLS-1$
                        ktoauszug.getJahr(), ktoauszug.getNummer(), ktoauszug.getErstellungsdatum()));
                if (AbstractMahoeSynchronizeJobKontoauszugPdf.isDoubleEntry(ktoauszug)) {
                    Logger.info(I18N.tr("AbstractMahoeSynchronizeJobKontoauszugPdf.accountStatementAlreaydSaved")); //$NON-NLS-1$
                } else {
                    if (Boolean.parseBoolean(MetaKey.KONTOAUSZUG_MARK_READ.get(this.konto))) {
                        ktoauszug.setGelesenAm(now);
                    }

                    ktoauszug.store();
                    KontoauszugPdfUtil.receive(ktoauszug, kontoauszugPDF.getData());

                    Application.getMessagingFactory().sendMessage(new ImportMessage(ktoauszug)); // NOPMD
                    this.konto.addToProtokoll(
                            I18N.tr("AbstractMahoeSynchronizeJobKontoauszugPdf.accountStatementsDownloaded"), //$NON-NLS-1$
                            Protokoll.TYP_SUCCESS);

                }

            }

            // Save download date -2

            MetaKey.KONTOAUSZUG_INTERVAL_LAST.set(this.konto, HBCI.LONGDATEFORMAT.format(new Date()));
            this.monitor.setPercentComplete(90);

        } catch (final RemoteException e) {
            Logger.error(I18N.tr("AbstractMahoeSynchronizeJobKontoauszugPdf.savePdfFailed"), e); //$NON-NLS-1$
            throw new ApplicationException(e.getMessage(), e);
        }
    }

    /**
     * Meldet das Webseiten-Objekt an der Bank an.
     *
     * @throws WebSiteException
     * @throws ApplicationException
     */
    protected abstract void login() throws WebSiteException, ApplicationException;

    /**
     * Gets all account sheets since last query.
     *
     * @return a list of found account statements as @link MahoeKontoauszugPDF
     * @throws WebSiteException
     * @throws ApplicationException
     */
    private List<MahoeKontoauszugPDF> getAccoutStatements() throws WebSiteException, ApplicationException {

        List<MahoeKontoauszugPDF> statements = new ArrayList<>();

        try {

            Logger.info(I18N.tr("AbstractMahoeSynchronizeJobKontoauszugPdf.lastDownload", //$NON-NLS-1$
                    MetaKey.KONTOAUSZUG_INTERVAL_LAST.get(this.konto)));
            statements = this.webSite.getAccountSheets(this.accountInfo);

        } catch (final RemoteException e) {
            throw new ApplicationException(I18N.tr("AbstractMahoeSynchronizeJobKontoauszugPdf.errorGettingStatements") //$NON-NLS-1$
                    , e);
        }

        return statements;
    }

}
