/*******************************************************************************
 * Copyright 2015 - 2019 fmCreation
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald diese von der
 * Europischen Kommission genehmigt wurden - Folgeversionen der EUPL
 * ("Lizenz"); Sie drfen dieses Werk ausschlielich gem dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 *   https://progs.frankmuenster.de/eupl_v1-2_de
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften gefordert oder in
 * schriftlicher Form vereinbart, wird die unter der Lizenz verbreitete Software
 * "so wie sie ist", OHNE JEGLICHE GEWHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrcklich oder stillschweigend - verbreitet. Die sprachspezifischen
 * Genehmigungen und Beschrnkungen unter der Lizenz sind dem Lizenztext zu
 * entnehmen.
 ******************************************************************************/

package de.frankmuenster.mahoe.utils;

import java.util.Arrays;
import java.util.List;

import org.apache.commons.lang3.Range;

/**
 * Kreditkartennummer &Uuml;berpr&uuml;fer. Er unterst&uuml;tzt im Moment die folgenden Karten:<br/>
 * MasterCard(incl. neue Version 2017), VISA und American Express (mit Anfangziffern 34 und 37).
 *
 * @author Frank Mnster
 */
public final class DebitCardValidator {

    /**
     * &Uuml;berpr&uuml;ft eine Kreditkartennumer auf G&uuml;ltigkeit.
     *
     * @param cardNumber
     * @return <i>true</i> wenn Kartennummer g&uuml;ltig ist, <i>false</i> sonst.
     */
    public static boolean isValidateCardNumber(final String cardNumber) {

        boolean valid = false;
        final DebitCardType cardType = DebitCardValidator.getCardType(cardNumber);
        if (cardType != null) {

            final List<Integer> allowedLength = cardType.getLength();
            if (allowedLength.contains(cardNumber.length())) {
                valid = DebitCardValidator.lunTest(cardNumber, cardType);
            }
        }

        return valid;
    }

    /**
     * Ermittelt den Kartentyp anhand der IIN
     *
     * @param cardNumber
     * @return den Kartentyp
     */
    private static DebitCardType getCardType(final String cardNumber) {

        DebitCardType cardType = null;

        for (final DebitCardType type : DebitCardType.values()) {

            final int iinLength = type.getIin().getMaximum().toString().length();
            final Integer iinToTest = Integer.valueOf(cardNumber.substring(0, iinLength));

            if (type.getIin().contains(iinToTest)) {
                cardType = type;
                break;
            }
        }

        return cardType;

    }

    /**
     * Berechnet die Quersumme einer bergebenen Zahl.
     *
     * @param digit
     *            die Zahl
     * @return die Quersumme
     */
    private static int getCrossTotal(final int digit) {

        int summe = 0;
        int internalDigit = digit;
        while (0 != internalDigit) {
            // addiere die letzte Ziffer
            summe = summe + internalDigit % 10;
            // entferne die letzte Ziffer
            internalDigit = internalDigit / 10;
        }
        return summe;
    }

    /**
     * F&uuml;hrt den LUN-Test aus.
     *
     * @param cardNumber
     * @param cardTyp
     * @return <i>true</i> wenn Kartennummer g&uuml;ltig ist, <i>false</i> sonst.
     */
    private static boolean lunTest(final String cardNumber, final DebitCardType cardTyp) {

        int querSumme = 0;
        final int start = cardTyp.usePrefix ? 0 : 1;
        for (int i = cardNumber.length() - 2; i >= start; i -= 2) {
            final Integer ziffer = cardNumber.charAt(i) - '0'; // Hart :-)
            querSumme += DebitCardValidator.getCrossTotal(ziffer * 2);
        }
        for (int i = cardNumber.length() - 1; i >= start; i -= 2) {
            final Integer ziffer = cardNumber.charAt(i) - '0';
            querSumme += DebitCardValidator.getCrossTotal(ziffer);
        }

        return querSumme % 10 == 0;
    }

    private DebitCardValidator() {

    }

    /**
     * Enum mit den unterst&uuml;tzenden Kreditkarten.
     *
     * @author Frank Mnster
     */
    private enum DebitCardType {

        MASTERCARD("MasterCard", Range.between(51, 55), new Integer[] { 16 }, true), //$NON-NLS-1$

        MASTERCARD2017("MasterCard", Range.between(2221, 2720), new Integer[] { 16 }, true), //$NON-NLS-1$

        VISA("VISA", Range.between(4, 4), new Integer[] { 13, 16, 19 }, true), //$NON-NLS-1$

        AMEX34("American Express", Range.between(34, 34), new Integer[] { 15 }, true), //$NON-NLS-1$

        AMEX37("American Express", Range.between(37, 37), new Integer[] { 15 }, true); //$NON-NLS-1$

        /**
         * Die IIN
         */
        private Range<Integer> iin;

        /**
         * Die erlaubten L&auml;ngen
         */
        private Integer[] length;

        /**
         * Der Name
         */
        private String name;

        /**
         * Benutze Prefixnummer in LUN-Test
         */
        private boolean usePrefix;

        DebitCardType(final String name, final Range<Integer> iin, final Integer[] length, final boolean usePrefix) { // NOPMD 

            this.name = name;
            this.iin = iin;
            this.length = length.clone();
            this.usePrefix = usePrefix;

        }

        /**
         * Liefert die IIN
         *
         * @return die IIN
         */
        public Range<Integer> getIin() {

            return this.iin;
        }

        /**
         * Liefert die erlaubten L&auml;ngen
         *
         * @return die erlaubten L&auml;ngen
         */
        public List<Integer> getLength() {

            return Arrays.asList(this.length);
        }

        /**
         * Liefert den Kreditkarteninstitusnamen
         *
         * @return den Kreditkarteninstitusnamen
         */
        @SuppressWarnings("unused")
        public String getName() {

            return this.name;
        }

        /**
         * Soll das Prefix beim LUN-TEST benutzt werden.
         *
         * @return Prefix beim LUN-TEST nutzen.
         */
        @SuppressWarnings("unused")
        public boolean isUsePrefix() {

            return this.usePrefix;
        }

    }

}
